require 'net/http'
require 'parsedate'
require 'md5'
require 'timeout'

class Site
  DIFF_THRESHOLD = 5

  module Status
    HEAD = :HEAD
    GET = :GET
    LENGTH = :LENGTH
    CACHE = :CACHE
    ERROR = '0'.intern
    TIMEOUT = :TIMEOUT
  end

  module Error
    ERROR = :ERROR
    TIMEOUT = :TIMEOUT
    LOOKAHEAD = :LOOKAHEAD
  end

  attr_accessor :url, :target_url, :title, :author, :category, :password,
    :last_modified, :last_detected, :last_visited, :content_length,
    :md5digest, :md5digest_line, :status, :error, :expires, :comment

  HTML_NEWLINE = /(\r|\n|<br\/?>)+/i

  class RetryURL < Exception; end

  def log(*args)
    Utility.log(*args.collect{|arg| "#{@url}: #{arg}"})
  end

  def valid_modified
    @last_modified && @expires.to_i > Time.now.to_i ? @last_modified : nil
  end

  def modify_specific_url
    if ! Setting::MODIFY_SPECIFIC_URL || ! @last_modified
      @url
    elsif @url.include?('?')
      @url.sub('\?', Time.at(@last_modified).strftime('?rrkdmy=%m%d%H%M&'))
    else
      @url + Time.at(@last_modified).strftime('?%m%d%H%M')
    end
  end

  def describe(output)
    output << "URL:    #{@url}\n"
    output << "TARGET: #{@target_url}\n" unless @url == @target_url
    output << "TITLE:  #{@title}\n"
    output << "AUTHOR: #{@author}\n"
    output << "SIZE:   #{@content_length}\n" if @content_length
    output << "L-M:    #{Time.at(@last_modified).rfc822fmt}\n" if @last_modified
    output << "L-M-D:  #{Time.at(@last_detected).rfc822fmt}\n" if @last_detected
    output << "EXPIRE: #{Time.at(@expires).rfc822fmt}\n" if @expires
    output << "STATUS: #{@status.id2name}\n" if @status
    output << "ERROR:  #{@error.id2name}\n" if @error
  end

  def detect(last_modified, status)
    if last_modified && 
        last_modified > Time.now.to_i + Setting::REFRESH_INTERVAL
      log("lookahead: #{last_modified}")
      @status = status
      @error = Error::LOOKAHEAD
      return
    end
    @last_modified = last_modified
    @status = status
    @last_detected = Time.now.to_i
    @expires = @last_detected + Setting::CACHE_EXPIRE_TIME
  end

  def update
    log('begin')

    @error = nil
    http = response = nil
    url = @target_url

    begin
      host, port, path = Utility.parse_url(url)

      timeout(Setting::HTTP_TIMEOUT) do 
        http = Net::HTTP.new(host, port)
      end

      methods = [ :get2 ]
      methods.unshift(:head2) if @status == Status::HEAD

      methods.each do |method|
        options = {}
        options['User-Agent'] = Ruriko::USER_AGENT
      
        if @last_modified
          options['If-Modified-Since'] = Time.at(@last_modified).utc.rfc822fmt
        end

        timeout(Setting::HTTP_TIMEOUT) do 
          response = http.send(method, path, options)
        end

        if response.key?('location')
          log('retry ya')
          raise RetryURL, response['location']
        elsif response.key?('last-modified')
          date = ParseDate.parsedate(response['last-modified'])
          time = Time.gm(*date[0, 6]).to_i
          detect(time, Status::HEAD)
          return
        elsif response.code == '304'
          detect(@last_modified, @status)
          return
        elsif response.code != '200'
          @error = Error::ERROR
          return
        end
      end
    rescue RetryURL => e
#      http.finish
      url = e.message
      retry
    end

    body = response.body

    @content_length = body.size

    body = Kconv.toeuc(body)

    Plugin.clippers.each do |clipper|
      next unless clipper.match(self)
      body = clipper.clip(body)
    end

    Plugin.parsers.each do |parser|
      next unless parser.match(self)
      last_modified = parser.parse(body)
      if last_modified
        detect(last_modified, Status::GET)
        return
      end
    end

    md5 = MD5.new(body)
    if md5.digest == @md5digest
      detect(@last_modified, Status::CACHE)
      return
    end
    @md5digest = md5.digest

    if response['content-type'] =~ /^text\/html/
      md5digest_line =
        body.split(HTML_NEWLINE).collect{|l| MD5.new(l).digest[0,2]}.join
    
      if @md5digest_line &&
          md5digest_line.size == @md5digest_line.size
        diff = count = 0
        (0..(md5digest_line.size)).each do |i|
          count += 1
          diff +=1 if md5digest_line[i] != @md5digest_line[i]
        end
        #      log("diff rate: #{diff * 100 / count}")
        return if diff * 100 / count < DIFF_THRESHOLD
      end
      
      @md5digest_line = md5digest_line
    end

    unless @last_detected
      @last_detected = Time.now.to_i
      return
    end

    detect((@last_visited + Time.now.to_i) / 2, Status::LENGTH)

  rescue TimeoutError
    @error = Error::TIMEOUT
    raise

  rescue
    @error = Error::ERROR
    raise

  ensure
    @last_visited = Time.now.to_i
    log('end')
#    http.finish
  end
end



